# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plkit']

package_data = \
{'': ['*']}

install_requires = \
['pytorch-lightning>=0.8,<0.9', 'scipy>=1.4,<2.0', 'sklearn', 'torch>=1.5,<2.0']

setup_kwargs = {
    'name': 'plkit',
    'version': '0.0.4',
    'description': 'Deep learning boilerplates on pytorch-lightning',
    'long_description': '# plkit\n\nDeep learning boilerplate based on [pytorch-lightning][1]\n\n## Installation\n```\npip install plkit\n```\n\n## Usage\n\n### From pytorch-lightning\'s minimal example\n\n```python\nfrom plkit import Module, Data, Trainner\n\nclass MINISTData(Data):\n\n    def data_reader(self, sources):\n        minst = MNIST(sources, train=True, download=True,\n                      transform=transforms.ToTensor())\n        return {\'train\': (minst.data, minst.targets)}\n\nclass LitClassifier(Module):\n\n    def __init__(self, config, num_classes, optim=\'adam\', loss=\'auto\'):\n        super().__init__(config, num_classes, optim, loss)\n        self.l1 = torch.nn.Linear(28 * 28, 10)\n\n    def forward(self, x):\n        return torch.relu(self.l1(x.view(x.size(0), -1)))\n\n    def training_step(self, batch, batch_nb):\n        x, y = batch\n        loss = self.loss_function(self(x), y)\n        tensorboard_logs = {\'train_loss\': loss}\n        return {\'loss\': loss, \'log\': tensorboard_logs}\n\n    # we don\'t need configure_optimizers anymore\n\nconfig = {\n    \'batch_size\': 32,\n    \'gpus\': 8,\n    \'precision\': 16,\n    \'num_classes\': 2,\n    # other options to initial a pytorch-lightning Trainer\n}\n\ndef main():\n    data = MINISTData(os.getcwd(), config[\'batch_size\'])\n    model = LitClassifier(config)\n    trainer = Trainer.from_dict(config, data=data)\n    trainer.fit(model)\n\nif __name__ == \'__main__\':\n    main()\n```\n\nOr even simpler for `main`:\n```python\nfrom plkit import run\n\n# Data class and model class definition\n# ...\n\ndef main():\n    run(config, LitClassifier, MINISTData)\n\n```\n\n## Features\n- Exposed terminal logger\n- Even more abstracted boilerplace (What you only need to care is your data and model)\n- Trainer from a dictionary (not only from an `ArgumentParser` or a `Namespace`)\n- Abstraction of data manager\n- Easy hyperparameter logging\n- Running `test` automatically when `test_dataloader` is given\n- Auto loss function and optimizer\n- Builtin measurements (using `pytorch-lightning.metrics`)\n\n## Exposed terminal logger\n```python\nfrom plkit import logger\n\n# You logic\n\nif __name__ == "__main__":\n    logger.info(\'Fantastic starts ...\')\n    # ...\n    logger.info(\'Pipeline done.\')\n```\n\n## More abstracted boilerplate\n```python\nfrom plkit import Module, Data, run\n\n# Data manager\nclass MyData(Data):\n\n    def data_reader(self, sources):\n        """Read your data from sources\n        Return a list of data\n        """\n        # your logic\n\n# Model\nclass MyModel(Module):\n    # Your model definition\n\nif __name__ == "__main__":\n    config = {\n        batch_size=32,\n        # other configs\n    }\n    run(config, MyModel, MyData)\n```\n\n## Trainer from a configuration dictionary\n\nApart from `Trainer.from_argparse_args` to create a trainer, we also added a way to create a trainer by `Trainer.from_dict`, which enables possibilities to use other argument parser packages. For example:\n```python\nfrom pyparam import params\nparams.gpus = 1\nparams.gpus.desc = \'Number of GPUs to use.\'\n\nconfig = params._parse()\n\ntrainer = Trainer.from_dict(config)\n# ...\n```\n\nWhy not using `**config` to create a trainer?\n```python\ntrainer = Trainer(**config)\n```\n\nThis is because sometimes `config` will have config items other than the arguments that `Trainer` constructor needs, which will raise an error. `Trainer.from_dict` filters the items only the `Trainer` needs.\n\n## Data manager\n\nAs it showed in the above examples, you don\'t need to concern about `Dataset` and `Dataloader` stuff. What you only need to care about is how to read the data from the sources. `plkit` will take care of the `Dataset` and `Dataloader` for you.\n\nUsually, you need to return a list of `data` (features) as well as the corresponding `targets` (labels) from your `data_reader` function. In such a case, you will not be able to fetch the names of the samples by `data, targets = batch` in the `training_step`, `validation_step` and `test_setp`.\n\nTo do that, you will need to provide `with_name=True` argument when instantialize `Data`: `data = Data(..., with_name=True)`, or in your config:\n```python\nconfig = {data_with_name=True, ...}\ndata = Data(sources, config)\n```\nand then, return dictionaries of data and targets with keys as the names. Finally, you are able to fetch the names in any of the steps by `data, targets, names = batch`.\n\n`plkit` can also split your data into `training`, `validation` and `test` parts, just by passing a ratio to `Data`: `data = Data(..., ratio=(.7, .2, .1))` (training: 70%, validation: 20%, test: 10%). Or in config: `config = {train_val_test_ratio: (.7, .2, .1)}`\n\nIf you don\'t specify a ratio, you will need to return dictionaries from `data_reader` with keys `train`, `val` and `test` for the data assigned to each part.\n\n## hyperparameter logging\n\nTo log hyperparameters using raw `pytorch-lightning` will be painful. See [#1228][2].\n\nHere instead of all those tweakings, we only need to use a decorator:\n```python\nfrom plkit import log_hparams, Module\n\n# ...\nclass MyModel(Module):\n\n    @log_hparams\n    def validation_step(self, batch, batch_idx):\n        # do you logic\n        # return the logs as what you did with pytorch-lightning\n```\n\nKeep in mind that, to enable this, you have to keep the default logger. Since we switched default logger from tensorboard logger to `HyperparamsSummaryTensorBoardLogger` implemented in `trainer.py` of `plkit`, whose idea was borrowed from [here][3].\n\nTo custom a logger, you have to subclass `HyperparamsSummaryTensorBoardLogger`, or just use it:\n```python\nfrom plkit.trainer import HyperparamsSummaryTensorBoardLogger\n\ntrainer = Trainer(logger=HyperparamsSummaryTensorBoardLogger(...), ...)\n```\n\nThe only side effect of this is to have a `__hparams_placeholder__` metric to be logged, which is used to be inserted to identify the metrics logs together with the hyperparameters.\n\n## Auto loss function and optimizer\n\nA loss function will be initialized according to the `optim` and `loss` configurations.\n\n`optim` supports `adam` and `sgd`, corresponding to `torch.optim.Adam` and `torch.optim.SGD`, respectively. You can specify `learning_rate` in the configuration\n\nIf `loss` is `auto`, `MSELoss` will be used for `num_classes==1` and `CrossEntropyLoss` otherwise. You can also specify the loss function by `loss=L1Loss()`.\n\n## Builtin measurements\n\nYou can get some of the measurements directly now by `self.measure(logits, labels, method, **kwargs)`, which calls metrics implemented in `pytorch_lightning.metrics`.\n\n- For `num_classes == 1` (regression), `mse`, `rmse`, `mae` and `rmsle` are available.\n- Otherwise, `accuracy`, `precision`, `recall`, `f1_score` and `iou`.\n\nFor extra `kwargs`, check the [source code][4] (Haven\'t found them documented yet).\n\n\n\n\n[1]: https://github.com/PyTorchLightning/pytorch-lightning\n[2]: https://github.com/PyTorchLightning/pytorch-lightning/issues/1228\n[3]: https://github.com/mRcSchwering/pytorch_lightning_test/blob/master/src/loggers.py\n[4]: https://github.com/PyTorchLightning/pytorch-lightning/tree/master/pytorch_lightning/metrics\n\n\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/plkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
